@@ -972,7 +972,10 @@ pango_attr_list_change (PangoAttrList  *list,
   g_return_if_fail (list != NULL);
 
   if (start_index == end_index)	/* empty, nothing to do */
-    return;
+    {
+      pango_attribute_destroy (attr);
+      return;
+    }
   
   tmp_list = list->attributes;
   prev = NULL;
@@ -1403,7 +1406,7 @@ pango_attr_iterator_get (PangoAttrIterator *iterator,
  *        so if you plan to keep it around, you must call:
  *        <literal>pango_font_description_set_family (desc, pango_font_description_get_family (desc))</literal>.
  * @language: if non-%NULL, location to store language tag for item, or %NULL
- *            if non is found.
+ *            if none is found.
  * @extra_attrs: if non-%NULL, location in which to store a list of non-font
  *           attributes at the the current position; only the highest priority
  *           value of each attribute will be added to this list. In order
