@@ -167,12 +167,12 @@ pango_context_set_font_map (PangoContext *context,
   g_return_if_fail (PANGO_IS_CONTEXT (context));
   g_return_if_fail (!font_map || PANGO_IS_FONT_MAP (font_map));
 
-  if (context->font_map)
-    g_object_unref (context->font_map);
-
   if (font_map)
     g_object_ref (font_map);
 
+  if (context->font_map)
+    g_object_unref (context->font_map);
+
   context->font_map = font_map;
 }
 
@@ -570,6 +570,9 @@ update_attr_iterator (ItemizeState *state)
 				&state->lang, &state->extra_attrs);
   state->copy_extra_attrs = FALSE;
 
+  if (!state->lang)
+    state->lang = state->context->language;
+  
   fallback = find_attribute (state->extra_attrs, PANGO_ATTR_FALLBACK);
   state->enable_fallback = (fallback == NULL || ((PangoAttrInt *)fallback)->value);
 
