@@ -28,8 +28,6 @@
 #include <glib.h>
 #include <glib/gprintf.h>
 
-#include <freetype/freetype.h>
-
 #include "pangoft2.h"
 #include "pangoft2-private.h"
 #include "pangofc-fontmap.h"
@@ -112,21 +110,6 @@ _pango_ft2_font_new (PangoFT2FontMap *ft2fontmap,
   return ft2font;
 }
 
-static gboolean
-set_unicode_charmap (FT_Face face)
-{
-  int charmap;
-
-  for (charmap = 0; charmap < face->num_charmaps; charmap++)
-    if (face->charmaps[charmap]->encoding == ft_encoding_unicode)
-      {
-	FT_Error error = FT_Set_Charmap(face, face->charmaps[charmap]);
-	return error == FT_Err_Ok;
-      }
-
-  return FALSE;
-}
-
 static void
 load_fallback_face (PangoFT2Font *ft2font,
 		    const char   *original_file)
@@ -171,12 +154,6 @@ load_fallback_face (PangoFT2Font *ft2font,
       g_free (name);
     }
 
-  if (!set_unicode_charmap (ft2font->face))
-    {
-      g_warning ("Unable to load unicode charmap from file %s, exiting\n", filename2);
-      exit (1);
-    }
-  
   FcPatternDestroy (sans);
   FcPatternDestroy (matched);
 }
@@ -188,8 +165,6 @@ load_fallback_face (PangoFT2Font *ft2font,
  * Returns the native FreeType2 FT_Face structure used for this PangoFont.
  * This may be useful if you want to use FreeType2 functions directly.
  *
- * Use pango_fc_font_get_face() instead.
- * 
  * Return value: a pointer to a #FT_Face structure, with the size set correctly
  **/
 FT_Face
@@ -250,17 +225,6 @@ pango_ft2_font_get_face (PangoFont *font)
 	}
 
       g_assert (ft2font->face);
-      
-      if (!set_unicode_charmap (ft2font->face))
-	{
-	  g_warning ("Unable to load unicode charmap from font file %s",
-                     filename);
-	  
-	  FT_Done_Face (ft2font->face);
-	  ft2font->face = NULL;
-	  
-	  load_fallback_face (ft2font, filename);
-	}
 
       error = FT_Set_Char_Size (ft2font->face,
 				PANGO_PIXELS_26_6 (ft2font->size),
@@ -654,14 +618,15 @@ static gboolean
 pango_ft2_font_real_has_char (PangoFcFont *font,
 			      gunichar     wc)
 {
-  FT_Face face;
-  FT_UInt index;
+  FcCharSet *charset;
 
-  face = pango_ft2_font_get_face ((PangoFont *)font);
-  index = FT_Get_Char_Index (face, wc);
-  return (index && index <= face->num_glyphs);
+  if (FcPatternGetCharSet (font->font_pattern,
+                           FC_CHARSET, 0, &charset) != FcResultMatch)
+    return FALSE;
+
+  return FcCharSetHasChar (charset, wc);
 }
-				   
+
 static guint
 pango_ft2_font_real_get_glyph (PangoFcFont *font,
 			       gunichar     wc)
@@ -670,7 +635,7 @@ pango_ft2_font_real_get_glyph (PangoFcFont *font,
   FT_UInt index;
 
   face = pango_ft2_font_get_face ((PangoFont *)font);
-  index = FT_Get_Char_Index (face, wc);
+  index = FcFreeTypeCharIndex (face, wc);
   if (index && index <= face->num_glyphs)
     return index;
 
@@ -840,6 +805,29 @@ pango_ft2_render_layout_line (FT_Bitmap       *bitmap,
 				x + PANGO_PIXELS (x_off + ink_rect.x),
 				x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width));
 	  break;
+	case PANGO_UNDERLINE_ERROR:
+          {
+            int point_x;
+            int counter = 0;
+	    int end_x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
+
+            for (point_x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
+                 point_x <= end_x;
+                 point_x += 2)
+	      {
+		if (counter)
+		  pango_ft2_draw_hline (bitmap,
+					risen_y + 2,
+					point_x, MIN (point_x + 1, end_x));
+		else
+		  pango_ft2_draw_hline (bitmap,
+					risen_y + 3,
+					point_x, MIN (point_x + 1, end_x));
+		
+		counter = (counter + 1) % 2;
+	      }
+          }
+	  break;
 	case PANGO_UNDERLINE_LOW:
 	  pango_ft2_draw_hline (bitmap,
 				risen_y + PANGO_PIXELS (ink_rect.y + ink_rect.height),
